package com.flyfish.framework.mongodb.query.impl;

import com.flyfish.framework.query.NamedQueryChain;
import com.flyfish.framework.query.Queries;
import com.flyfish.framework.query.QueryChain;
import com.flyfish.framework.query.QueryCondition;
import com.flyfish.framework.query.holder.QueryChainHolder;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.bson.types.ObjectId;
import org.springframework.data.mongodb.core.query.Criteria;

import java.util.Collection;
import java.util.List;
import java.util.function.Supplier;

/**
 * mongodb实现的基于名称的查询链
 * 这里需要一个机制，能够将父类上下文带入，并且能够知晓实际查询对象的连接方式
 * 这就需要用到观察者模式，以便延迟执行。将动作作为事件发布到对象里，在具体的执行时机再进行fire。
 * 于是乎，我们便可以优雅而且低耦合的实现链式了。
 *
 * @author wangyu
 */
public class MongoNamedQueryChain extends MongoQueryDefinition implements NamedQueryChain {

    /**
     * 以且连接下一个字段
     *
     * @param column 列
     * @return 查询条件
     */
    @Override
    public QueryCondition<NamedQueryChain> and(String column) {
        and();
        return new NamedQueryCondition(column);
    }

    /**
     * 直接拼接提供者，此处懒加载，最终build才会执行
     *
     * @param supplier 提供者
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> NamedQueryChain and(Supplier<QueryChain<V, ?>> supplier) {
        and().with(() -> supplier.get().build());
        return this;
    }

    /**
     * 条件列表
     *
     * @param chain 多个条件们
     * @return 结果
     */
    @Override
    public NamedQueryChain and(NamedQueryChain chain) {
        and().with(chain::build);
        return this;
    }

    /**
     * 多个嵌套子条件列表
     *
     * @param chains     多条链
     * @param combinator 各个链条之间的连接方式
     * @return 结果
     */
    @Override
    public NamedQueryChain and(List<NamedQueryChain> chains, Queries.Combinator combinator) {
        return and(combine(chains, combinator));
    }

    /**
     * 以或连接下一个字段
     *
     * @param column 列
     * @return 查询条件
     */
    @Override
    public QueryCondition<NamedQueryChain> or(String column) {
        or();
        return new NamedQueryCondition(column);
    }

    /**
     * 以或连接条件列表
     *
     * @param chain 多个条件们
     * @return 结果
     */
    @Override
    public NamedQueryChain or(NamedQueryChain chain) {
        or().with(chain::build);
        return this;
    }

    /**
     * 以或连接嵌套多个子条件列表
     *
     * @param chains     多条链
     * @param combinator 各个链条之间的连接方式
     * @return 结果
     */
    @Override
    public NamedQueryChain or(List<NamedQueryChain> chains, Queries.Combinator combinator) {
        return or(combine(chains, combinator));
    }

    /**
     * 直接拼接提供者，此处懒加载，最终build才会执行
     *
     * @param supplier 提供者
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> NamedQueryChain or(Supplier<QueryChain<V, ?>> supplier) {
        or().with(() -> supplier.get().build());
        return this;
    }

    /**
     * 以且的关系进行拼接
     *
     * @return 结果
     */
    private QueryChainHolder<Criteria> and() {
        return holder.link(Criteria::andOperator);
    }

    /**
     * 以或的关系进行拼接
     *
     * @return 结果
     */
    private QueryChainHolder<Criteria> or() {
        return holder.link(Criteria::orOperator);
    }

    /**
     * 结合多个查询链，并根据组合方式进行处理
     *
     * @param chains     查询链集合
     * @param combinator 组合方式
     * @return 结果
     */
    private NamedQueryChain combine(List<NamedQueryChain> chains, Queries.Combinator combinator) {
        return chains.stream().reduce((left, right) -> combinator == Queries.Combinator.OR ? left.or(right) : left.and(right))
                .orElse(new MongoNamedQueryChain());
    }


    /**
     * 条件操作实现
     *
     * @author wangyu
     */
    @RequiredArgsConstructor
    class NamedQueryCondition implements QueryCondition<NamedQueryChain> {

        private final String column;

        /**
         * 相等判定
         *
         * @param value 值
         * @return 结果
         */
        @Override
        public NamedQueryChain eq(Object value) {
            if (null != value) {
                holder.with(Criteria.where(column).is(value));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 判定某列的值中存在指定值，特指json array数据类型且子类型中带有id的场景
         * 用于兼容mongodb查询，mysql查询使用JSON_CONTAINS进行判定。
         * 当且仅当mongodb会拼接.$id，mysql一律匹配id字段
         * <p>
         * 等价于 .eq(value)
         *
         * @param value 值
         * @return 结果
         */
        @Override
        public NamedQueryChain hasId(Object value) {
            if (null != value) {
                holder.with(Criteria.where(column + ".$id").is(new ObjectId(String.valueOf(value))));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 对于mongodb，自动处理。对于关系型数据库，代表json array中是否包含对应值，值仅支持基本数据类型
         * <p>
         * 等价于 .eq(value)
         *
         * @param value 基本数据类型的值
         * @return 结果
         */
        @Override
        public NamedQueryChain has(Object value) {
            if (null != value) {
                holder.with(Criteria.where(column).is(value));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 不等判定
         *
         * @param value 值
         * @return 结果
         */
        @Override
        public NamedQueryChain ne(Object value) {
            if (null != value) {
                holder.with(Criteria.where(column).ne(value));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 值介于两者之间
         *
         * @param items 双值列表
         * @return 结果
         */
        @Override
        public NamedQueryChain between(List<?> items) {
            if (CollectionUtils.size(items) == 2) {
                holder.with(Criteria.where(column).gte(items.get(0)).lte(items.get(1)));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 模糊匹配，这里是全模糊
         *
         * @param keyword 查询关键字
         * @return 结果
         */
        @Override
        public NamedQueryChain like(String keyword) {
            if (StringUtils.isNotBlank(keyword)) {
                holder.with(Criteria.where(column).regex(keyword));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 根据指定的方向进行模糊查询
         *
         * @param keyword   关键字
         * @param direction 方向，可以匹配开头和结尾
         * @return 结果
         */
        @Override
        public NamedQueryChain like(String keyword, Queries.Direction direction) {
            if (StringUtils.isNotBlank(keyword)) {
                // 匹配开头
                if (direction == Queries.Direction.LEFT) {
                    keyword = "^" + keyword;
                } else if (direction == Queries.Direction.RIGHT) {
                    keyword = keyword + "$";
                }
                holder.with(Criteria.where(column).regex(keyword));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 判定为空
         *
         * @return 结果
         */
        @Override
        public NamedQueryChain isNull() {
            holder.with(Criteria.where(column).isNull());
            return MongoNamedQueryChain.this;
        }

        /**
         * 包含在内
         *
         * @param list 集合
         * @return 结果
         */
        @Override
        public NamedQueryChain in(Collection<?> list) {
            if (CollectionUtils.isNotEmpty(list)) {
                holder.with(Criteria.where(column).in(list));
            }
            return MongoNamedQueryChain.this;
        }

        /**
         * 包含在内
         * 注意，需要根据字段类型推断。
         * 如果是json数组，需要进行双向匹配
         *
         * @param values 值们
         * @return 结果
         */
        @Override
        public NamedQueryChain in(Object... values) {
            if (ArrayUtils.isNotEmpty(values)) {
                holder.with(Criteria.where(column).in(values));
            }
            return MongoNamedQueryChain.this;
        }
    }
}
